import subprocess
from tchotcho.log import log
from typing import Optional
import pathlib
import attr
import click


@attr.s(auto_attribs=True)
class ProcMsg:
    cmd: str = attr.ib()
    stdout: str = attr.ib()
    stderr: str = attr.ib()
    ok: bool = attr.ib()
    code: Optional[int] = attr.ib()


class ShellManager(object):
    def run_cmd(self, cmd, _iter=False):
        log.info("Executing cmd: %s" % " ".join(cmd))

        proc = subprocess.Popen(
            cmd,
            shell=False,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
        if not _iter:
            try:
                stdout, stderr = proc.communicate()
            except subprocess.TimeoutExpired:
                proc.kill()
                stdout, stderr = proc.communicate()
            except Exception:
                proc.kill()
                proc.wait()
        else:
            stdout = proc.stdout
            stderr = proc.stderr

        returncode = proc.poll()

        log.debug(proc.args)

        pmsg = ProcMsg(cmd, stdout, stderr, returncode == 0, returncode)

        if not pmsg.ok:
            log.error("%s returncode: %s stderr: %s" % (cmd[0], pmsg.code, pmsg.stderr))
        else:
            log.info(f"{cmd[0]} ok: {pmsg.ok} returncode: {pmsg.code}")
        return pmsg

    def rsync(self, src, dst, dry=False, privat_key=None):
        if not any([char in src for char in (":", "@")]):
            src = str(pathlib.Path(src).resolve())

        if not any([char in dst for char in (":", "@")]):
            dst = str(pathlib.Path(dst).resolve())
        option = "-avzm"
        option += "n" if dry else ""
        cmd = [
            "rsync",
            option,
            # precalculate file list
            "--no-inc-recursive",
            # get nicer summary
            "--info=progress2",
            "--delete",
            "--exclude",
            ".git",
            "--filter=:- .gitignore",
        ]
        if privat_key:
            cmd.extend(["-e", f"ssh -o StrictHostKeyChecking=no -i {privat_key}"])
        cmd.extend([src, dst])
        pmsg = self.run_cmd(cmd)
        if pmsg.ok:
            summary = " ".join([x for x in pmsg.stdout.split("\n") if x][-2:])
            log.info(f"rsync summary: {summary}")
        return pmsg

    def ssh(self, privat_key, host, command, _iter):
        cmd = [
            "ssh",
            "-i",
            privat_key,
            "-o",
            "ServerAliveInterval=60",
            "-o",
            "ServerAliveCountMax=2",
            "-o",
            "StrictHostKeyChecking=no",
            host,
            command,
        ]
        pmsg = self.run_cmd(cmd, _iter=_iter)
        if pmsg.ok:
            log.info(f"ssh stdout: {pmsg.stdout}")
        return pmsg


mgr = ShellManager()


@click.group()
def shell():
    ...


@shell.command()
@click.option(
    "--privat-key",
    help="Path to key file",
    required=True,
    type=click.Path(exists=True, resolve_path=True),
)
@click.option("--host", help="Host url e.g.: (user@server)", required=True)
@click.option(
    "--cmd", help="Command to execute via ssh e.g.: (pwd ; ls -l)", required=True
)
@click.option("--it/--no-it", default=False)
def ssh(privat_key, host, cmd, it):
    pmsg = mgr.ssh(privat_key, host, cmd, it)
    if it:
        for line in pmsg.stdout:
            click.echo(line)
    else:
        if pmsg.ok:
            msg = pmsg.stdout
        else:
            msg = pmsg.stderr
        click.echo(msg)


@shell.command()
@click.option("--src", help="Source path", required=True)
@click.option(
    "--dst", help="Destination path e.g.: (/home/... or user@server)", required=True
)
@click.option("--dry/--no-dry", help="Only print yaml no create", default=False)
@click.option(
    "--privat-key",
    help="Path to key file",
    type=click.Path(exists=True, resolve_path=True),
)
def rsync(src, dst, dry, privat_key):
    pmsg = mgr.rsync(src, dst, dry, privat_key)
    if pmsg.ok:
        msg = " ".join([x for x in pmsg.stdout.split("\n") if x][-2:])
    else:
        msg = pmsg.stderr
    click.echo(msg)



import json
import datetime
import itertools
import collections.abc
import concurrent.futures

import boto3
import click
import colorama
import pandas as pd

import tabulate

from tchotcho.config import get_settings

colorama.init()


class SpotManager(object):
    def __init__(self):
        self.settings = get_settings()

    def _spot_history_wrapper(self, args):
        return self._spot_history(*args)

    def _spot_history(self, inst, region):
        client = boto3.client("ec2", region_name=region)
        prices = client.describe_spot_price_history(
            InstanceTypes=inst,
            ProductDescriptions=["Linux/UNIX", "Linux/UNIX (Amazon VPC)"],
            StartTime=(
                datetime.datetime.now() + datetime.timedelta(days=1)
            ).isoformat(),
            MaxResults=len(inst),
        )
        ret = prices["SpotPriceHistory"]
        return ret

    def list(self, gpu, region=None, inst=[], method="future-process"):
        if len(inst) == 0:
            with open(self.settings.GPU_INFO_FILE) as f:
                inst = json.load(f)["instance"]
                if gpu:
                    inst = [x["name"] for x in inst if x["gpu"] > 0 and x["supported"]]
                else:
                    inst = [x["name"] for x in inst]

        if not isinstance(inst, collections.abc.Sequence) or isinstance(inst, str):
            raise Exception("We need a list for inst!")

        client = boto3.client("ec2")
        regions = [
            (inst, x["RegionName"]) for x in client.describe_regions()["Regions"]
        ]

        if region:
            regions = [x for x in regions if region in x[1]]

        if method == "future-process":
            with concurrent.futures.ProcessPoolExecutor() as executor:
                results = list(executor.map(self._spot_history_wrapper, regions))
        else:
            results = [self._spot_history(*x) for x in regions]

        results = [x for x in results if x]

        results = list(itertools.chain.from_iterable(results))

        for x in results:
            x["SpotPrice"] = float(x["SpotPrice"])
        ret = sorted(results, key=lambda x: x["SpotPrice"])
        print(f"Found {len(results)} entries")
        return ret


mgr = SpotManager()


@click.group()
def spot():
    ...


@spot.command(name="list")
@click.option("--region", help="List spot prices in region")
@click.option(
    "--gpu/--no-gpu", default=True, help="Limit results to show only GPU instances"
)
@click.option("--inst", help="List spot prices for instance type", multiple=True)
@click.option("--csv/--no-csv", default=False)
def _list(gpu, region, inst, csv):
    ret = mgr.list(gpu, region, inst)

    def set_color(val):
        if val < 1:
            val = colorama.Back.GREEN + str(val) + colorama.Back.RESET
        elif val > 3:
            val = colorama.Back.RED + str(val) + colorama.Back.RESET
        else:
            val = colorama.Back.YELLOW + str(val) + colorama.Back.RESET
        return val

    # apply to specific column
    df = pd.DataFrame(ret)
    df = df[["InstanceType", "AvailabilityZone", "SpotPrice"]]
    to_print = df.to_csv()

    if not csv:
        df["SpotPrice"] = df["SpotPrice"].apply(set_color)
        to_print = tabulate.tabulate(
            df, headers="keys", tablefmt="fancy_grid", showindex="never"
        )
    print(to_print)



import boto3
import os
import troposphere.iam
import troposphere.ec2
import troposphere
import troposphere.cloudformation
import textwrap
import awacs.aws
import awacs.sts
import awacs


def get_default_vpc_id():
    """Returns a default VPC id or None"""
    ec2 = boto3.client("ec2")
    res = ec2.describe_vpcs(Filters=[{"Name": "isDefault", "Values": ["true"]}])
    if not len(res["Vpcs"]):
        return None
    return res["Vpcs"][0]["VpcId"]


def create_cloudformation(
    key_name,
    ami_id,
    instance_type,
    security_group=None,
    subnet_id=None,
    price=None,
    size=100,
    extra_user_data="",
):

    # XXX set this to get real bool values
    os.environ["TROPO_REAL_BOOL"] = "true"

    t = troposphere.Template(Description="TchoTcho EC2 train")

    instance_security_group = t.add_resource(
        troposphere.ec2.SecurityGroup(
            "InstanceSecurityGroup",
            VpcId=get_default_vpc_id(),
            GroupDescription="Enable only SSH ingoing via port 22 and all outgoing",
            SecurityGroupIngress=[
                troposphere.ec2.SecurityGroupRule(
                    IpProtocol="tcp", FromPort=22, ToPort=22, CidrIp="0.0.0.0/0"
                ),
                troposphere.ec2.SecurityGroupRule(
                    IpProtocol="tcp", FromPort=22, ToPort=22, CidrIpv6="::/0"
                ),
            ],
            SecurityGroupEgress=[
                troposphere.ec2.SecurityGroupRule(IpProtocol="-1", CidrIp="0.0.0.0/0"),
                troposphere.ec2.SecurityGroupRule(IpProtocol="-1", CidrIpv6="::/0"),
            ],
        )
    )

    instance_role = t.add_resource(
        troposphere.iam.Role(
            "InstanceRole",
            AssumeRolePolicyDocument=awacs.aws.Policy(
                Version="2012-10-17",
                Statement=[
                    awacs.aws.Statement(
                        Effect=awacs.aws.Allow,
                        Principal=awacs.aws.Principal("Service", "ec2.amazonaws.com"),
                        Action=[awacs.sts.AssumeRole],
                    ),
                ],
            ),
            ManagedPolicyArns=[
                "arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role",
            ],
            Policies=[
                troposphere.iam.Policy(
                    PolicyName="S3FullAccess",
                    PolicyDocument={
                        "Statement": [
                            {"Effect": "Allow", "Action": "s3:*", "Resource": "*"}
                        ],
                    },
                )
            ],
        )
    )

    instance_profile = t.add_resource(
        troposphere.iam.InstanceProfile(
            "InstanceProfile", Roles=[troposphere.Ref(instance_role)],
        )
    )

    launch_template = t.add_resource(
        troposphere.ec2.LaunchTemplate(
            "InstanceLaunchTemplate",
            # https://github.com/cloudtools/troposphere/blob/07dde6b66fca28dd401903027d8ac13bc107e0b6/examples/CloudFormation_Init_ConfigSet.py#L45
            # https://stackoverflow.com/questions/35095950/what-are-the-benefits-of-cfn-init-over-userdata
            # XXX for now we are not using this we always just delete the stack
            # Metadata=troposphere.cloudformation.Metadata(
            #     troposphere.cloudformation.Init(
            #         troposphere.cloudformation.InitConfigSets(default=["bootstrap"]),
            #         awspackages=troposphere.cloudformation.InitConfig(
            #             commands={
            #                 "001-bootstrap": {"command": "touch ~/bootstra.txt"},
            #             },
            #         ),
            #     ),
            # ),
            LaunchTemplateData=troposphere.ec2.LaunchTemplateData(
                KeyName=key_name,
                ImageId=ami_id,
                InstanceType=instance_type,
                UserData=troposphere.Base64(
                    # Sub is only needed if we have variables
                    # troposphere.Sub(
                    textwrap.dedent(
                        """#!/bin/bash
                        set -x -e

                        # remove ubuntu daily updates
                        # https://unix.stackexchange.com/a/480986
                        systemd-run --property="After=apt-daily.service apt-daily-upgrade.service" --wait /bin/true
                        systemctl mask apt-daily.service apt-daily-upgrade.service

                        apt update
                        apt install neovim silversearcher-ag -y

                        # extra user data
                        {extra_user_data}
                        """.format(
                            extra_user_data=extra_user_data
                        )
                    ),
                    # )
                ),
                IamInstanceProfile=troposphere.ec2.IamInstanceProfile(
                    Arn=troposphere.GetAtt(instance_profile, "Arn"),
                ),
                BlockDeviceMappings=[
                    troposphere.ec2.LaunchTemplateBlockDeviceMapping(
                        DeviceName="/dev/sda1",
                        Ebs=troposphere.ec2.EBSBlockDevice(
                            DeleteOnTermination=True, VolumeSize=size, Encrypted=True
                        ),
                    )
                ],
            ),
        )
    )

    if price:
        instance_market_options = troposphere.ec2.InstanceMarketOptions(
            MarketType="spot",
            SpotOptions=troposphere.ec2.SpotOptions(
                SpotInstanceType="one-time",
                MaxPrice=str(price),
                InstanceInterruptionBehavior="terminate",
            ),
        )

        launch_template.properties["LaunchTemplateData"].properties[
            "InstanceMarketOptions"
        ] = instance_market_options

    if not security_group:
        security_group = [troposphere.Ref(instance_security_group)]

    if subnet_id:
        network_interfaces = [
            troposphere.ec2.NetworkInterfaces(
                SubnetId=subnet_id, DeviceIndex=0, Groups=[security_group],
            )
        ]

        launch_template.properties["LaunchTemplateData"].properties[
            "NetworkInterfaces"
        ] = network_interfaces
    else:
        launch_template.properties["LaunchTemplateData"].properties[
            "SecurityGroupIds"
        ] = [troposphere.Ref(instance_security_group)]

    ec2_instance = t.add_resource(
        troposphere.ec2.Instance(
            "Instance",
            LaunchTemplate=troposphere.ec2.LaunchTemplateSpecification(
                LaunchTemplateId=troposphere.Ref(launch_template),
                Version=troposphere.GetAtt(launch_template, "LatestVersionNumber"),
            ),
        )
    )

    t.add_output(
        [
            troposphere.Output(
                "InstanceId",
                Description="InstanceId of the EC2 instance",
                Value=troposphere.Ref(ec2_instance),
            ),
            troposphere.Output(
                "AZ",
                Description="Availability Zone of the EC2 instance",
                Value=troposphere.GetAtt(ec2_instance, "AvailabilityZone"),
            ),
            troposphere.Output(
                "PublicIP",
                Description="Public IP address of the EC2 instance",
                Value=troposphere.GetAtt(ec2_instance, "PublicIp"),
            ),
            troposphere.Output(
                "PrivateIP",
                Description="Private IP address of the EC2 instance",
                Value=troposphere.GetAtt(ec2_instance, "PrivateIp"),
            ),
            troposphere.Output(
                "PublicDNS",
                Description="Public DNSName of the EC2 instance",
                Value=troposphere.GetAtt(ec2_instance, "PublicDnsName"),
            ),
            troposphere.Output(
                "PrivateDNS",
                Description="Private DNSName of the EC2 instance",
                Value=troposphere.GetAtt(ec2_instance, "PrivateDnsName"),
            ),
        ]
    )
    # XXX moto has some problems with yaml; validate, LaunchTemplateData is
    # not parsed so the ec2instance ImageId other keys are not found
    # return t.to_yaml()
    return t.to_json()



import boto3
from tchotcho.tropo import create_cloudformation
from tchotcho.util import boto_exception, get_wrapped_waiter
from tchotcho.log import log
import pandas as pd
import click
import colorama

import tabulate

colorama.init()


class StackManager(object):
    def __init__(self):
        self.cf = boto3.client("cloudformation")
        self.output = None

    def _parse_template(self, template_data):
        self.cf.validate_template(TemplateBody=template_data)
        return template_data

    def stack_exists(self, name):
        stacks = self.cf.list_stacks()["StackSummaries"]
        for stack in stacks:
            if stack["StackStatus"] == "DELETE_COMPLETE":
                continue
            if name == stack["StackName"]:
                return True
        return False

    def _waiter_callback(self, response):
        log.info(response)
        if "Stacks" in response:
            log.info(response["Stacks"])
            if "Outputs" in response["Stacks"][0]:
                out = response["Stacks"][0]["Outputs"]
                df = pd.DataFrame(out)
                self.output = {x["OutputKey"].lower(): x["OutputValue"] for x in out}
                to_print = tabulate.tabulate(
                    df, headers="keys", tablefmt="fancy_grid", showindex="never"
                )
                print(to_print)
        else:
            log.info(response)

    @boto_exception
    def delete(self, name):
        """Delete a stack if it exists by name """
        log.info("Stack is: %s", name)

        if self.stack_exists(name):
            log.info(f"Deleting stack: {name}...")
            self.cf.delete_stack(StackName=name)
            waiter = get_wrapped_waiter(
                self.cf, "stack_delete_complete", self._waiter_callback
            )
            waiter.wait(StackName=name)
            log.info(f"Stack {name} deleted")
        else:
            log.error(f"Stack {name} does not exist.")

    @boto_exception
    def create(
        self,
        name,
        ami,
        inst,
        security_group,
        subnet,
        price,
        size,
        dry,
        extra_user_data='echo "hello" > /tmp/hello.txt',
    ):
        "Create stack"

        template = create_cloudformation(
            name,
            ami,
            inst,
            security_group,
            subnet,
            price,
            size,
            extra_user_data,
        )

        template_data = self._parse_template(template)
        if dry:
            return template_data

        params = {
            "StackName": name,
            "TemplateBody": template_data,
            "Capabilities": ["CAPABILITY_IAM"],
        }

        if self.stack_exists(name):
            log.error(f"Stack {name} already exists!")
        else:
            print(f"Creating stack: {name}...")
            self.cf.create_stack(**params)
            waiter = get_wrapped_waiter(
                self.cf, "stack_create_complete", self._waiter_callback
            )
            waiter.wait(StackName=name)
            log.info(f"Stack {name} created")
            return self.output

    @boto_exception
    def list(self):
        resp = self.cf.describe_stacks()
        return resp


mgr = StackManager()


@click.group()
def stack():
    ...


@stack.command()
@click.option("--name", required=True, help="Name of stack to create && key")
@click.option(
    "--ami",
    required=True,
    help="Name of ami to use",
    default="ami-062a3145bcf312c71",
    show_default=True,
)
@click.option("--inst", required=True, help="Name of the instance to use")
@click.option("--security_group", help="Name of the security group to use")
@click.option("--subnet", help="Name of the subnet to use")
@click.option("--price", type=float, help="Name of the instance to use")
@click.option("--size", type=int, help="Size of the disk in GB", default=120)
@click.option("--dry/--no-dry", help="Only print yaml no create", default=False)
def create(name, ami, inst, security_group, subnet, price, size, dry):
    ret = mgr.create(name, ami, inst, security_group, subnet, price, size, dry)
    ret = click.echo(ret)
    click.echo(ret)


@stack.command()
@click.option("--name", required=True, help="Name of stack to delete")
def delete(name):
    mgr.delete(name)


@stack.command()
@click.option("--name", help="List stacks by name")
@click.option("--csv/--no-csv", default=False)
def list(name, csv):
    ret = mgr.list()
    stacks = ret.get("Stacks")
    if not stacks:
        click.echo("No stacks found!")
        return

    ret = ret["Stacks"]

    def set_color(val):
        if val == name:
            val = colorama.Back.GREEN + val + colorama.Back.RESET
        return val

    # apply to specific column
    df = pd.DataFrame(ret)
    df = df[["StackName", "StackStatus", "CreationTime", "Capabilities"]]
    to_print = df.to_json()

    if not csv:
        df["StackName"] = df["StackName"].apply(set_color)
        to_print = tabulate.tabulate(
            df, headers="keys", tablefmt="fancy_grid", showindex="never"
        )
    print(to_print)



from tchotcho.action.key import key as key_group
from tchotcho.action.stack import stack as stack_group
from tchotcho.action.spot import spot as spot_group
from tchotcho.action.info import info as info_group
from tchotcho.action.shell import shell as shell_group
import click


@click.group()
def cli():
    ...


cli.add_command(key_group)
cli.add_command(stack_group)
cli.add_command(spot_group)
cli.add_command(info_group)
cli.add_command(shell_group)
